/*
 *************************************************************************
 *                                                                       *
 * Copyright 2000 Compaq Computer Corporation                            *
 *                                                                       *
 * COMPAQ Registered in U.S. Patent and Trademark Office.                *
 *                                                                       *
 *************************************************************************
 * IMPORTANT: Carefully read the License Terms below before              *
 * proceeding.  By use of these materials you agree to these terms.      *
 * If you do not agree to these terms, you may not use this software or  *
 * the accompanying documentation.                                       *
 *************************************************************************
 * LICENSE TERMS                                                         *
 * 1. GRANT                                                              *
 * Compaq Computer Corporation ("COMPAQ") grants you the right to use,   *
 * modify, and distribute the following source code (the "Software")     *
 * on any number of computers. You may use the Software as part of       *
 * creating a software program or product intended for commercial or     *
 * non-commercial distribution in machine-readable source code, binary,  *
 * or executable formats. You may distribute the Software as             *
 * machine-readable source code provided this license is not removed     *
 * from the Software and any modifications are conspicuously indicated.  *
 * 2. COPYRIGHT                                                          *
 * The Software is owned by COMPAQ and its suppliers and is protected by *
 * copyright laws and international treaties.  Your use of the Software  *
 * and associated documentation is subject to the applicable copyright   *
 * laws and the express rights and restrictions of these terms.          *
 * 3. RESTRICTIONS                                                       *
 * You may not remove any copyright, trademark, or other proprietary     *
 * notices from the Software or the associated  documentation.           *
 * You are responsible for compliance with all applicable export or      *
 * re-export control laws and regulations if you export the Software.    *
 * This license is governed by and is to be construed under the laws     *
 * of the State of Texas.                                                *
 *                                                                       *
 * DISCLAIMER OF WARRANTY AND LIABILITY                                  *
 * Compaq shall not be liable for technical or editorial errors or       *
 * omissions contained herein. The information contained herein is       *
 * subject to change without notice.                                     *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.       *
 * THE ENTIRE RISK ARISING OUT OF THE USE OF THIS SOFTWARE REMAINS WITH  *
 * RECIPIENT.  IN NO EVENT SHALL COMPAQ BE LIABLE FOR ANY DIRECT,        *
 * CONSEQUENTIAL, INCIDENTAL, SPECIAL, PUNITIVE OR OTHER DAMAGES         *
 * WHATSOEVER (INCLUDING WITHOUT LIMITATION DAMAGES FOR LOSS OF BUSINESS *
 * PROFITS, BUSINESS INTERRUPTION, OR LOSS OF BUSINESS INFORMATION),     *
 * EVEN IF COMPAQ HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES    *
 * AND WHETHER IN AN ACTION OF CONTRACT OR TORT INCLUDING NEGLIGENCE.    *
 *                                                                       *
 * If you have any questions concerning this license, please contact:    *
 * Compaq Computer Corporation, Software Business Practices, ZKO1-2/D22, *
 * 110 Spit Brook Road, Nashua, NH. 03062-2698.                          *
 *                                                                       *
 *************************************************************************
 */

#include "vms_jackets.h"
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <file.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <stat.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef __ALPHA
#include <gen64def.h>
#else
#include "gen64def.h"
#endif
#include <dcdef.h>
#include <descrip.h>
#include <starlet.h>
#include <stsdef.h>
#include <devdef.h>
#include <dvidef.h>
#include <dvsdef.h>
#include <lib$routines.h>

/*
** This module contains a set of jackets for the opendir set of routines.
**
** opendir()
** changed so that if a regular opendir fails, then we:
**   - check to see if its "/" - if so, start a device scan and return that
**   - check to see if it looks like a device only spec. If it does, then
**     we retry the opendir with 000000 appended.
**
** readdir()
** changed to check the pointer passed in. If its the special opendir root
** pointer then return the next device from a device scan.
**
** rewinddir()
** changed to check the pointer passed in. If its the special opendir root
** pointer then reset the device scan search.
**
** closedir()
** changed to check the pointer passed in. If its the special opendir root
** pointer then close off the device scan stuff.
**
** seekdir()
** telldir()
** changed to check the pointer passed in. If its the special opendir root
** pointer then return a pointer for our readdir context.



/**********************************************************************/
/*  Jacket routines to handle VMS and decc limits                     */
/*                                                                    */
/*  Routine: GENERIC_OPENDIR_JACKET                                   */
/*                                                                    */
/*  Jacket for: opendir                                               */
/**********************************************************************/


/*
** Calls in here to getcwd need to go to the jacketted version.
*/
#define getcwd GENERIC_GETCWD_JACKET

/*
** This must be in a supplied header file some place, but I never been
** able to find it.
*/
typedef struct {
    short   buflen;
    short   itmcode;
    void    *buffer;
    void    *retlen;
} ITMLST;

/*
** This is the context that we use for calling device_scan. It contains the
** context that's used by device_scan itself, plus the dirent that we
** return back to the caller (this way we support multiple concurrent
** readdir's on the root directory.
*/
typedef struct {
    struct _generic_64 devscanctx;
    struct _generic_64 saved_devscanctx;
    struct dirent de;
} DIRCTX;

/*
** We only want to getenv logicals once.
*/
char *stat_device_roots = (char *) -1;

/*
** This isn't in VMS_JACKETS.H because if it were, we'd need to include
** stat.h in there, and that causes other problems since there is a
** routine "stat" and a struct "stat".
*/
int GENERIC_STAT_JACKET(const char *path, struct stat *buf);

DIR *GENERIC_OPENDIR_JACKET2(const char *path) {

    int len;
    char my_path[PATH_MAX+8];
    DIR *rv;

    /*
    ** Check that we have the room to do our special stuff.
    */
    len=strlen(path);
    if (len <= PATH_MAX) {

        /*
        ** First off, tidy the spec up a little.
        */
        generic_tidy_spec(my_path, path);

        /*
        ** If its the root directory, then we initialize a device scan
        ** context that in GENERIC_readdir_Jacket will cause all disk device
        ** names to be returned. We add one to the address of our own context
        ** so that we can recosnize it in readdir, closedir, etc. Yes, this is
        ** a little hacky but the real opendir always returns pointers to
        ** objects that are at least word-aligned.
        */
        if ( (strlen(my_path) == 1) && (my_path[0] == '/') ) {
            DIRCTX *ctx;
            ctx = calloc(1,sizeof(DIRCTX));
            if (!ctx) {
                errno = ENOMEM;
                return NULL;
            }
            return (DIR*)((int)ctx+1);
        }

        /*
        ** If its an absolute spec with only one term, then try appending
        ** 000000 and see if that works. See comment in GENERIC_STAT_JACKET
        ** for why we have to do this now, before we call the real opendir.
        */
        if (generic_absone(my_path)) {
            len = strlen(my_path);
            if (my_path[len-1] == '/')
                my_path[len-1] = '\0';
            strcat(my_path,"/000000");
            rv=opendir(my_path);
            if (rv)
                return rv;
        }
    }

    /*
    ** Nothing special, so fall thru to a regular opendir, but on the
    ** tidied spec.
    */
    return opendir(my_path);
}

struct dirent *GENERIC_READDIR_JACKET(DIR *dir) {

    struct dsc$descriptor_s devnam;
    DIRCTX *ctx;
    char devnam_buf[PATH_MAX+7+1]; /* allow for "/000000" */
    unsigned short retlen;
    ITMLST itmlst[2];
    static int devclass = DC$_DISK;
    int rv;

    /*
    ** If its not one of ours, then pass on to the real readdir. The only
    ** change we make is to strip off the training "." of any file spec
    ** that doesn't have a file type.
    */
    if (!((int)dir & 1)) {
        struct dirent *de;
        de=readdir(dir);
        if (de) {
            char *p=de->d_name;
            int l=strlen(p);
            if (p[l-1] == '.')
                p[l-1] = '\0';
            }
        return de;
    }

    /*
    ** This is one of ours. Get our context.
    */
    ctx = (DIRCTX *)((int)dir-1);

    /*
    ** We need to make another $DEVICE_SCAN call.
    */

    itmlst[0].buflen  = sizeof(devclass);
    itmlst[0].itmcode = DVS$_DEVCLASS;
    itmlst[0].buffer  = &devclass;
    itmlst[0].retlen  = NULL;

    itmlst[1].buflen  = 0;
    itmlst[1].itmcode = 0;
    itmlst[1].buffer  = NULL;
    itmlst[1].retlen  = NULL;

    devnam.dsc$b_dtype=DSC$K_DTYPE_T;
    devnam.dsc$b_class=DSC$K_CLASS_S;
    devnam.dsc$w_length=PATH_MAX;
    devnam.dsc$a_pointer=devnam_buf;

    /*
    ** Do the device searching in a loop, because we don't accept
    ** just any disk and we might need to make several calls to
    ** $device_scan before we find a disk we can return to the caller.
    */

    while (1) {

        rv=sys$device_scan (
                        &devnam,        /* returned device name */
                        &retlen,        /* returned name length */
                        NULL,           /* search devnam */
                        &itmlst,        /* itemlist */
                        &ctx->devscanctx/* context */
                      );
        if ($VMS_STATUS_SUCCESS(rv)) {
            /*
            ** We got a device, but we only way ones which are MNT and AVL.
            */
            int devchar, item=DVI$_DEVCHAR;

            /*
            ** Call GETDVI to get the device characteristics to we can
            ** determine if it makes sense to return this one to the caller.
            ** Note that we mess with the descriptor length and we must
            ** remember to put it back in case we have to make another
            ** device_scan call.
            */
            devnam.dsc$w_length=retlen;
            rv=lib$getdvi(&item,0,&devnam,&devchar,NULL,NULL);
            devnam.dsc$w_length=PATH_MAX;

            if (!$VMS_STATUS_SUCCESS(rv))
                continue; /* if we can't getdvi it, ignore it */

            if ( (devchar & DEV$M_MNT) && (devchar & DEV$M_AVL) ) {
                /*
                ** This device has all the characteristics we want, and so
                ** we can return this device to the caller. Remember that
                ** the caller is calling readdir and so we have to package
                ** the device name up to look like a direct.
                */
                struct dirent *de = &ctx->de;
                struct stat st;
                char *pin ,*pout, sbuf[PATH_MAX+2];
                int i;

                /*
                ** Copy the device name to the dirent structure, lowercasing
                ** it on the way. We also omit any leading underscore and
                ** trailing colon.
                */
                pin=devnam_buf;

                if (pin[retlen-1] == ':')
                    pin[retlen-1] = '\0';
                else
                    pin[retlen] = '\0';

                if (*pin == '_')
                    *pin++;

                pout=de->d_name;
                do {
                    *pout++=tolower(*pin);
                } while (*pin++);

                /*
                ** Get the fileid for the root directory of this device.
                ** Logically we want the device to be its root directory so
                ** this is probably the most sensible thing to return.
                **
                ** If for some reason the stat fails, then there's no point
                ** in returning this device to the caller, since when he
                ** stats it he'll get the same error.
                **
                ** stat'ing certain device roots can be REAL slow (eg. an
                ** inforserved device can take 15-20 seconds to respond).
                ** So, we take a short cut and assume that the FID will
                ** be 4,4,0. Unless STAT_DEVICE_ROOTS is set in which case
                ** we will do a real stat.
                */
                if (stat_device_roots == (char *) -1)
                    stat_device_roots = getenv(VMS_PREFIX "STAT_DEVICE_ROOTS");

                if (stat_device_roots) {
                    sbuf[0]='/';
                    strcpy(&sbuf[1],de->d_name);
                    if (GENERIC_STAT_JACKET(sbuf,&st))
                        continue;

                    de->d_ino[0] = st.st_ino[0];
                    de->d_ino[1] = st.st_ino[1];
                    de->d_ino[2] = st.st_ino[2];
                }
                else {
                    de->d_ino[0] = 4;
                    de->d_ino[1] = 4;
                    de->d_ino[2] = 0;
                }

                /*
                ** All set. Return this dirent describing the device name.
                */
                return de;
            }
        }
        else
            /*
            ** Some error in the device_scan call. It was probably "no more
            ** devices". Assume so, and tell the caller "no more".
            */
            return NULL;
    }
}
                        
void GENERIC_REWINDDIR_JACKET(DIR *dir) {

    DIRCTX *ctx;

    /*
    ** If its not one of ours, then pass on to the real closedir.
    */
    if (!((int)dir & 1)) {
        rewinddir(dir);
        return;
    }

    /*
    ** This is one of ours. Get our context.
    */
    ctx = (DIRCTX *)((int)dir-1);

    /*
    ** To rewind the readdir of the root directory we just have
    ** to set the device_scan context to zero. This will cause
    ** the next call to device_scan to start over. For simplicity,
    ** we just nuke our entire context area.
    */
    bzero(ctx,sizeof(DIRCTX));
}

int GENERIC_CLOSEDIR_JACKET(DIR *dir) {

    DIRCTX *ctx;

    /*
    ** If its not one of ours, then pass on to the real closedir.
    */
    if (!((int)dir & 1))
        return closedir(dir);

    /*
    ** This is one of ours. Get our context.
    */
    ctx = (DIRCTX *)((int)dir-1);

    /*
    ** Free the memory we used for our context and return.
    */
    free(ctx);
    return 0;
}

/*
** The context maintained by devicescan is 64 bits while the context we
** return back from telldir is only 32 bits. For now we'll just maintain
** one context pointer, and return an error (ENOBUFS) if asked for a
** second telldir. If this turns out to be a problem then we can fix it.
*/

long int GENERIC_TELLDIR_JACKET(DIR *dir) {

    DIRCTX *ctx;

    /*
    ** If its not one of ours, then pass on to the real telldir.
    */
    if (!((int)dir & 1))
        return telldir(dir);

    /*
    ** This is one of ours. Get our context.
    */
    ctx = (DIRCTX *)((int)dir-1);

    /*
    ** Save the current devscan context, and return an index (1 for now)
    ** to the caller.
    */
    if ( (ctx->saved_devscanctx.gen64$l_longword[0] == 0) &&
         (ctx->saved_devscanctx.gen64$l_longword[1] == 0) ) {
        memmove(&ctx->saved_devscanctx, &ctx->devscanctx,
                sizeof(ctx->devscanctx));
        return 1;
    }
    else {
        printf("ERROR - telldir is out of bufs\n");
        errno = ENOBUFS;
        return -1;
    }

}

void GENERIC_SEEKDIR_JACKET(DIR *dir, long int location) {

    DIRCTX *ctx;

    /*
    ** If its not one of ours, then pass on to the real seekdir.
    */
    if (!((int)dir & 1)) {
        seekdir(dir,location);
        return;
    }

    /*
    ** This is one of ours. Get our context.
    */
    ctx = (DIRCTX *)((int)dir-1);

    /*
    ** Revert back to the saved devscan context. We could do some sanity/error
    ** checking, but what's the point. We can't return any errors to the
    ** caller.
    */
    memmove(&ctx->devscanctx, &ctx->saved_devscanctx,
                sizeof(ctx->devscanctx));
    return;
}
                             



/*
** This is the normal file name checking jacket. This is called before
** the above opendir.
*/

DIR *GENERIC_OPENDIR_JACKET(const char *dir_spec)
{
   char temp_string[TEMP_STRING_SIZE];
   char *new_dir_spec;
   int status;

   new_dir_spec = GENERIC_VMS_Filename_parser((char *)dir_spec,
                                              temp_string,
					      GENERIC_K_STRING_TYPE_DIRECTORY,
                                              &status);
 
   if ((strlen(dir_spec) < 4) &&
        (!strcmp(dir_spec,".") ||
        !strcmp(dir_spec,"./") ||
        !strcmp(dir_spec,"<>") ||
        !strcmp(dir_spec,"[]") ||
        !strcmp(dir_spec,".//")))
   { /* for cwd */
        new_dir_spec = getcwd(temp_string,TEMP_STRING_SIZE-1);
   }

   GENERIC_TRACE("opendir_Jacket: Mapped:", dir_spec, "To:", new_dir_spec);

return(GENERIC_OPENDIR_JACKET2(new_dir_spec));
}
